/*
 * Copyright (c) 2014, 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Cmad unit tests.
 */

#include <limits.h>
#include <math.h>
#include <stdlib.h>

#include <cadef.h>
#include <epicsEvent.h>
#include <epicsMutex.h>
#include <epicsStdio.h>
#include <epicsUnitTest.h>
#include <testMain.h>

#include "cmad.h"
#include "cmad_internal.h"

#define AD_CAM1_PV_PREFIX   "13SIM1:cam1:"
#define AD_IMAGE1_PV_PREFIX "13SIM1:image1:"

#define AD_PV_NAME_ACQUIRE_PERIOD  AD_CAM1_PV_PREFIX   "AcquirePeriod"
#define AD_PV_NAME_ACQUIRE         AD_CAM1_PV_PREFIX   "Acquire"
#define AD_PV_NAME_ARRAY_CALLBACKS AD_CAM1_PV_PREFIX   "ArrayCallbacks"
#define AD_PV_NAME_SIZE_X          AD_CAM1_PV_PREFIX   "SizeX"
#define AD_PV_NAME_SIZE_Y          AD_CAM1_PV_PREFIX   "SizeY"
#define AD_PV_NAME_COLOR_MODE      AD_CAM1_PV_PREFIX   "ColorMode"
#define AD_PV_NAME_IMAGE1_ENABLE   AD_IMAGE1_PV_PREFIX "EnableCallbacks"

#define AD_COLOR_MODE_MONO 0
#define AD_COLOR_MODE_RGB1 1

#define CMAD_TEST_NUM_CHANNELS 7

#define CMAD_TEST_CH_ACQUIRE_PERIOD  0
#define CMAD_TEST_CH_ACQUIRE         1
#define CMAD_TEST_CH_ARRAY_CALLBACKS 2
#define CMAD_TEST_CH_SIZE_X          3
#define CMAD_TEST_CH_SIZE_Y          4
#define CMAD_TEST_CH_COLOR_MODE      5
#define CMAD_TEST_CH_IMAGE1_ENABLE   6

#define CMAD_TEST_PV_PREFIX    AD_IMAGE1_PV_PREFIX
#define CMAD_TEST_IMAGE_WIDTH  1024
#define CMAD_TEST_IMAGE_HEIGHT 576
#define CMAD_TEST_COLOR_MODE   AD_COLOR_MODE_RGB1
#define CMAD_TEST_FPS          60.0
#define CMAD_TEST_TIMEOUT      5.0

#define CMAD_TEST_CA_WORKER_ERR_MSG_SIZE 256

static struct cmad_client *client;

static epicsEventId ca_worker_do_eid;
static epicsEventId ca_worker_ack_eid;

static epicsMutexId ca_worker_mid;
static int ca_worker_stop;
static dbr_double_t ca_worker_acquire_period;
static dbr_enum_t ca_worker_acquire;
static dbr_enum_t ca_worker_array_callbacks;
static dbr_long_t ca_worker_size_x;
static dbr_long_t ca_worker_size_y;
static dbr_enum_t ca_worker_color_mode;
static dbr_enum_t ca_worker_image1_enable;
static void *ca_worker_value[CMAD_TEST_NUM_CHANNELS];
static int ca_worker_marked[CMAD_TEST_NUM_CHANNELS];
static char ca_worker_err_msg[CMAD_TEST_CA_WORKER_ERR_MSG_SIZE];

static int cmad_test_test_client_new(void);
static int cmad_test_test_client_new_bad_args(void);
static int cmad_test_test_client_free(void);
static int cmad_test_test_error(void);
static int cmad_test_test_is_connected(void);
static int cmad_test_test_is_ready_bad_width(void);
static int cmad_test_test_is_ready_bad_height(void);
static int cmad_test_test_is_ready_bad_color_mode(void);
static int cmad_test_test_is_ready_bad_to_good(void);
static int cmad_test_test_start_image_capture(void);
static int cmad_test_test_stop_image_capture(void);
static int cmad_test_test_capture_image(void);
static int cmad_test_test_capture_image_multiple(void);
static int cmad_test_test_capture_image_timeout(void);
static int cmad_test_test_capture_image_timeout_no_change(void);
static int cmad_test_test_capture_image_multiple_clients(void);
static int cmad_test_test_capture_image_quick(void);
static int cmad_test_test_capture_image_quick_multiple(void);
static int cmad_test_test_capture_image_quick_timeout(void);
static int cmad_test_test_capture_image_quick_timeout_no_change(void);
static int cmad_test_test_capture_image_quick_multiple_clients(void);

static void cmad_test_program_set_up(void);
static void cmad_test_program_tear_down(void);

static void cmad_test_set_up(void);
static void cmad_test_set_up_with_timeout(double);
static void cmad_test_tear_down(void);

static void cmad_test_reset_ca_worker_communication(void);
static void cmad_test_reset_ca_worker_communication_locked(void);

static void cmad_test_set_ca_worker_stop(int);
static void cmad_test_set_acquire_period(dbr_double_t);
static void cmad_test_set_acquire(dbr_enum_t);
static void cmad_test_set_array_callbacks(dbr_enum_t);
static void cmad_test_set_size_x(dbr_long_t);
static void cmad_test_set_size_y(dbr_long_t);
static void cmad_test_set_color_mode(dbr_enum_t);
static void cmad_test_set_image1_enable(dbr_enum_t);
static void cmad_test_set_defaults(void);

static int cmad_test_wait_connected(struct cmad_client *, int, double);
static int cmad_test_wait_ready(struct cmad_client *, int, double);

static void cmad_test_ca_worker_do(void);

static void cmad_test_ca_worker(void *);

MAIN(cmadTest)
{
  testPlan(21);

  cmad_test_program_set_up();

  testOk1(cmad_test_test_client_new());
  testOk1(cmad_test_test_client_new_bad_args());
  testOk1(cmad_test_test_client_free());
  testOk1(cmad_test_test_error());
  testOk1(cmad_test_test_is_connected());
  testOk1(cmad_test_test_is_ready_bad_width());
  testOk1(cmad_test_test_is_ready_bad_height());
  testOk1(cmad_test_test_is_ready_bad_color_mode());
  testOk1(cmad_test_test_is_ready_bad_to_good());
  testOk1(cmad_test_test_start_image_capture());
  testOk1(cmad_test_test_stop_image_capture());
  testOk1(cmad_test_test_capture_image());
  testOk1(cmad_test_test_capture_image_multiple());
  testOk1(cmad_test_test_capture_image_timeout());
  testOk1(cmad_test_test_capture_image_timeout_no_change());
  testOk1(cmad_test_test_capture_image_multiple_clients());
  testOk1(cmad_test_test_capture_image_quick());
  testOk1(cmad_test_test_capture_image_quick_multiple());
  testOk1(cmad_test_test_capture_image_quick_timeout());
  testOk1(cmad_test_test_capture_image_quick_timeout_no_change());
  testOk1(cmad_test_test_capture_image_quick_multiple_clients());

  cmad_test_program_tear_down();

  return testDone();
}

static int
cmad_test_test_client_new(void)
{
  int pass;

  pass = 0;

  client = cmad_client_new(CMAD_TEST_PV_PREFIX, CMAD_TEST_IMAGE_WIDTH,
      CMAD_TEST_IMAGE_HEIGHT, CMAD_TEST_TIMEOUT);
  if (client == NULL) goto err;
  cmad_client_free(client);

  pass = 1;

err:
  return pass;
}

static int
cmad_test_test_client_new_bad_args(void)
{
  int pass;

  pass = 0;

  client = cmad_client_new(NULL, CMAD_TEST_IMAGE_WIDTH,
      CMAD_TEST_IMAGE_HEIGHT, CMAD_TEST_TIMEOUT);
  if (client != NULL) goto err;
  client = cmad_client_new(CMAD_TEST_PV_PREFIX, 0,
      CMAD_TEST_IMAGE_HEIGHT, CMAD_TEST_TIMEOUT);
  if (client != NULL) goto err;
  client = cmad_client_new(CMAD_TEST_PV_PREFIX, CMAD_TEST_IMAGE_WIDTH,
      0, CMAD_TEST_TIMEOUT);
  if (client != NULL) goto err;
  client = cmad_client_new(CMAD_TEST_PV_PREFIX, ULONG_MAX,
      ULONG_MAX, CMAD_TEST_TIMEOUT);
  if (client != NULL) goto err;
  client = cmad_client_new(CMAD_TEST_PV_PREFIX, CMAD_TEST_IMAGE_WIDTH,
      CMAD_TEST_IMAGE_HEIGHT, 0.0);
  if (client != NULL) goto err;
  client = cmad_client_new(CMAD_TEST_PV_PREFIX, CMAD_TEST_IMAGE_WIDTH,
      CMAD_TEST_IMAGE_HEIGHT, -1.0);
  if (client != NULL) goto err;

  pass = 1;
  goto out;

err:
  cmad_client_free(client);

out:
  return pass;
}

static int
cmad_test_test_client_free(void)
{
  int pass;

  pass = 0;

  /* Ensure can free a client. */
  client = cmad_client_new(CMAD_TEST_PV_PREFIX, CMAD_TEST_IMAGE_WIDTH,
      CMAD_TEST_IMAGE_HEIGHT, CMAD_TEST_TIMEOUT);
  if (client == NULL) goto err;
  cmad_client_free(client);

  /* Ensure can cope with a null pointer. */
  cmad_client_free(NULL);

  pass = 1;

err:
  return pass;
}

static int
cmad_test_test_error(void)
{
  int pass;
  const char *err_msg;
  unsigned char buf[1];
  int status;

  pass = 0;
  err_msg = NULL;

  cmad_test_set_up();

  err_msg = cmad_error(client);
  if (err_msg != NULL) {
    testDiag("client should not have error message but does");
    goto err;
  }

  status = cmad_capture_image(client, buf, sizeof(buf));
  if (status != -1) {
    testDiag("capture should fail when buffer size != image size");
    goto err;
  }

  err_msg = cmad_error(client);
  if (err_msg == NULL) {
    testDiag("client should have error message about unsuccessful capture");
    goto err;
  }

  pass = 1;

err:
  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_is_connected(void)
{
  int pass;
  int status;

  pass = 0;

  cmad_test_set_up();

  status = cmad_test_wait_connected(client, 1, CMAD_TEST_TIMEOUT);
  if (status == -1) {
    testDiag("client never became ready");
    goto err;
  }

  pass = 1;

err:
  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_is_ready_bad_width(void)
{
  int pass;
  int status;

  pass = 0;

  cmad_test_set_up();

  status = cmad_test_wait_ready(client, 1, CMAD_TEST_TIMEOUT);
  if (status == -1) {
    testDiag("client never became ready");
    goto err;
  }

  cmad_test_set_size_x(CMAD_TEST_IMAGE_WIDTH - 1);
  cmad_test_ca_worker_do();

  status = cmad_test_wait_ready(client, 0, CMAD_TEST_TIMEOUT);
  if (status != 0) {
    testDiag("client never became not ready");
    goto err;
  }

  pass = 1;

err:
  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_is_ready_bad_height(void)
{
  int pass;
  int status;

  pass = 0;

  cmad_test_set_up();

  status = cmad_test_wait_ready(client, 1, CMAD_TEST_TIMEOUT);
  if (status == -1) {
    testDiag("client never became ready");
    goto err;
  }

  cmad_test_set_size_x(CMAD_TEST_IMAGE_HEIGHT - 1);
  cmad_test_ca_worker_do();

  status = cmad_test_wait_ready(client, 0, CMAD_TEST_TIMEOUT);
  if (status != 0) {
    testDiag("client never became not ready");
    goto err;
  }

  pass = 1;

err:
  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_is_ready_bad_color_mode(void)
{
  int pass;
  int status;

  pass = 0;

  cmad_test_set_up();

  status = cmad_test_wait_ready(client, 1, CMAD_TEST_TIMEOUT);
  if (status == -1) {
    testDiag("client never became ready");
    goto err;
  }

  cmad_test_set_color_mode(AD_COLOR_MODE_MONO);
  cmad_test_ca_worker_do();

  status = cmad_test_wait_ready(client, 0, CMAD_TEST_TIMEOUT);
  if (status != 0) {
    testDiag("client never became not ready");
    goto err;
  }

  pass = 1;

err:
  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_is_ready_bad_to_good(void)
{
  int pass;
  int status;

  pass = 0;

  cmad_test_set_up();

  status = cmad_test_wait_ready(client, 1, CMAD_TEST_TIMEOUT);
  if (status == -1) {
    testDiag("client never became ready");
    goto err;
  }

  cmad_test_set_color_mode(AD_COLOR_MODE_MONO);
  cmad_test_ca_worker_do();

  status = cmad_test_wait_ready(client, 0, CMAD_TEST_TIMEOUT);
  if (status != 0) {
    testDiag("client never became not ready");
    goto err;
  }

  cmad_test_reset_ca_worker_communication();
  cmad_test_set_color_mode(CMAD_TEST_COLOR_MODE);
  cmad_test_ca_worker_do();

  status = cmad_test_wait_ready(client, 1, CMAD_TEST_TIMEOUT);
  if (status == -1) {
    testDiag("client never became ready again");
    goto err;
  }

  pass = 1;

err:
  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_start_image_capture(void)
{
  int pass;
  int status;

  pass = 0;

  cmad_test_set_up();

  status = cmad_start_image_capture(client);
  if (status == -1) goto err;

  /* It's OK to start a capture that is already started. */
  status = cmad_start_image_capture(client);
  if (status == -1) goto err;

  pass = 1;

err:
  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_stop_image_capture(void)
{
  int pass;
  int status;

  pass = 0;

  cmad_test_set_up();

  status = cmad_start_image_capture(client);
  if (status == -1) goto err;

  status = cmad_stop_image_capture(client);
  if (status == -1) goto err;

  /* It's OK to stop a capture that is already stopped. */
  status = cmad_stop_image_capture(client);
  if (status == -1) goto err;

  pass = 1;

err:
  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_capture_image(void)
{
  int pass;
  unsigned char *buf;
  size_t buf_size;
  int status;

  pass = 0;
  buf = NULL;
  buf_size = CMAD_TEST_IMAGE_WIDTH * CMAD_TEST_IMAGE_HEIGHT * 3 * sizeof(*buf);

  cmad_test_set_up();

  status = cmad_start_image_capture(client);
  if (status == -1) {
    testDiag("unable to start image capture");
    goto err;
  }

  buf = malloc(buf_size);
  if (buf == NULL) {
    testDiag("unable to allocate image buffer");
    goto err;
  }

  status = cmad_capture_image(client, buf, buf_size);
  if (status == -1) {
    testDiag("unable to capture image: %s", cmad_error(client));
    goto err;
  }

  pass = 1;

err:
  if (buf != NULL) free(buf);

  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_capture_image_multiple(void)
{
  int i;
  int pass;
  unsigned char *buf;
  size_t buf_size;
  int status;

  i = 0;
  pass = 0;
  buf = NULL;
  buf_size = CMAD_TEST_IMAGE_WIDTH * CMAD_TEST_IMAGE_HEIGHT * 3 * sizeof(*buf);

  cmad_test_set_up();

  status = cmad_start_image_capture(client);
  if (status == -1) {
    testDiag("unable to start image capture");
    goto err;
  }

  buf = malloc(buf_size);
  if (buf == NULL) {
    testDiag("unable to allocate image buffer");
    goto err;
  }

  for (i = 0; i < 4; i++) {
    status = cmad_capture_image(client, buf, buf_size);
    if (status == -1) {
      testDiag("unable to capture image: %s", cmad_error(client));
      goto err;
    }
  }

  pass = 1;

err:
  if (buf != NULL) free(buf);

  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_capture_image_timeout(void)
{
  int pass;
  unsigned char *buf;
  size_t buf_size;
  int status;

  pass = 0;
  buf = NULL;
  buf_size = CMAD_TEST_IMAGE_WIDTH * CMAD_TEST_IMAGE_HEIGHT * 3 * sizeof(*buf);

  cmad_test_set_up_with_timeout(1.0);

  status = cmad_start_image_capture(client);
  if (status == -1) {
    testDiag("unable to start image capture");
    goto err;
  }

  cmad_test_set_color_mode(AD_COLOR_MODE_MONO);
  cmad_test_ca_worker_do();

  status = cmad_test_wait_ready(client, 0, CMAD_TEST_TIMEOUT);
  if (status != 0) {
    testDiag("client never became not ready");
    goto err;
  }

  buf = malloc(buf_size);
  if (buf == NULL) {
    testDiag("unable to allocate image buffer");
    goto err;
  }

  status = cmad_capture_image(client, buf, buf_size);
  if (status != -1) {
    testDiag("image capture should have timed out");
    goto err;
  }

  pass = 1;

err:
  if (buf != NULL) free(buf);

  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_capture_image_timeout_no_change(void)
{
  int pass;
  unsigned char *buf;
  size_t buf_size;
  int status;
  epicsTimeStamp start_time;
  epicsTimeStamp now_time;
  double elapsed_time;

  pass = 0;
  buf = NULL;
  buf_size = CMAD_TEST_IMAGE_WIDTH * CMAD_TEST_IMAGE_HEIGHT * 3 * sizeof(*buf);

  cmad_test_set_up();

  status = cmad_start_image_capture(client);
  if (status == -1) {
    testDiag("unable to start image capture");
    goto err;
  }

  buf = malloc(buf_size);
  if (buf == NULL) {
    testDiag("unable to allocate image buffer");
    goto err;
  }

  cmad_test_set_acquire_period(2.0);
  cmad_test_ca_worker_do();

  /* Synchronize with the 2.0 second acquire period. */
  for (;;) {
    status = epicsTimeGetCurrent(&start_time);
    if (status == -1) {
      testDiag("unable to obtain current time");
      goto err;
    }
    status = cmad_capture_image(client, buf, buf_size);
    if (status != 0) {
      testDiag("image capture should have succeeded");
      goto err;
    }
    status = epicsTimeGetCurrent(&now_time);
    if (status == -1) {
      testDiag("unable to obtain current time");
      goto err;
    }
    elapsed_time = epicsTimeDiffInSeconds(&now_time, &start_time);
    if (elapsed_time < 0.0) {
      /* Time advanced backward; consider now to be the new start time. */
      start_time = now_time;
    } else if (elapsed_time > 1.75) {
      /* Synchronization achieved. */
      break;
    }
  }

  status = cmad_set_timeout(client, 1.0);
  if (status == -1) {
    testDiag("unable to set timeout");
    goto err;
  }

  status = cmad_capture_image(client, buf, buf_size);
  if (status != -1) {
    testDiag("image capture should have timed out");
    goto err;
  }

  pass = 1;

err:
  if (buf != NULL) free(buf);

  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_capture_image_multiple_clients(void)
{
  int pass;
  unsigned char *buf;
  size_t buf_size;
  struct cmad_client *client2;
  int i;
  int j;
  int status;

  pass = 0;
  buf = NULL;
  buf_size = CMAD_TEST_IMAGE_WIDTH * CMAD_TEST_IMAGE_HEIGHT * 3 * sizeof(*buf);
  client2 = NULL;

  cmad_test_set_up();

  client2 = cmad_client_new(CMAD_TEST_PV_PREFIX, CMAD_TEST_IMAGE_WIDTH,
      CMAD_TEST_IMAGE_HEIGHT, CMAD_TEST_TIMEOUT);
  if (client2 == NULL) {
    testDiag("unable to create client2");
    goto err;
  }

  for (i = 0; i < 2; i++) {
    status = cmad_start_image_capture((i == 0) ? client : client2);
    if (status == -1) {
      testDiag("unable to start client%s image capture", (i == 0) ? "" : "2");
      goto err;
    }
  }

  buf = malloc(buf_size);
  if (buf == NULL) {
    testDiag("unable to allocate image buffer");
    goto err;
  }

  for (i = 0; i < 4; i++) {
    for (j = 0; j < 2; j++) {
      status = cmad_capture_image((j == 0) ? client : client2, buf, buf_size);
      if (status == -1) {
        testDiag("unable to capture client%s image: %s", (j == 0) ? "" : "2",
            cmad_error(client));
        goto err;
      }
    }
  }

  pass = 1;

err:
  if (buf != NULL) free(buf);
  if (client2 != NULL) cmad_client_free(client2);

  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_capture_image_quick(void)
{
  int pass;
  unsigned char *buf;
  size_t buf_size;
  int status;

  pass = 0;
  buf = NULL;
  buf_size = CMAD_TEST_IMAGE_WIDTH * CMAD_TEST_IMAGE_HEIGHT * 3 * sizeof(*buf);

  cmad_test_set_up();

  status = cmad_start_image_capture(client);
  if (status == -1) {
    testDiag("unable to start image capture");
    goto err;
  }

  buf = malloc(buf_size);
  if (buf == NULL) {
    testDiag("unable to allocate image buffer");
    goto err;
  }

  status = cmad_capture_image_quick(client, buf, buf_size);
  if (status == -1) {
    testDiag("unable to capture image: %s", cmad_error(client));
    goto err;
  }

  pass = 1;

err:
  if (buf != NULL) free(buf);

  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_capture_image_quick_multiple(void)
{
  int i;
  int pass;
  unsigned char *buf;
  size_t buf_size;
  int status;

  i = 0;
  pass = 0;
  buf = NULL;
  buf_size = CMAD_TEST_IMAGE_WIDTH * CMAD_TEST_IMAGE_HEIGHT * 3 * sizeof(*buf);

  cmad_test_set_up();

  status = cmad_start_image_capture(client);
  if (status == -1) {
    testDiag("unable to start image capture");
    goto err;
  }

  buf = malloc(buf_size);
  if (buf == NULL) {
    testDiag("unable to allocate image buffer");
    goto err;
  }

  for (i = 0; i < 4; i++) {
    status = cmad_capture_image_quick(client, buf, buf_size);
    if (status == -1) {
      testDiag("unable to capture image: %s", cmad_error(client));
      goto err;
    }
  }

  pass = 1;

err:
  if (buf != NULL) free(buf);

  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_capture_image_quick_timeout(void)
{
  int pass;
  unsigned char *buf;
  size_t buf_size;
  int status;

  pass = 0;
  buf = NULL;
  buf_size = CMAD_TEST_IMAGE_WIDTH * CMAD_TEST_IMAGE_HEIGHT * 3 * sizeof(*buf);

  cmad_test_set_up_with_timeout(1.0);

  status = cmad_start_image_capture(client);
  if (status == -1) {
    testDiag("unable to start image capture");
    goto err;
  }

  cmad_test_set_color_mode(AD_COLOR_MODE_MONO);
  cmad_test_ca_worker_do();

  status = cmad_test_wait_ready(client, 0, CMAD_TEST_TIMEOUT);
  if (status != 0) {
    testDiag("client never became not ready");
    goto err;
  }

  buf = malloc(buf_size);
  if (buf == NULL) {
    testDiag("unable to allocate image buffer");
    goto err;
  }

  status = cmad_capture_image_quick(client, buf, buf_size);
  if (status != -1) {
    testDiag("image capture should have timed out");
    goto err;
  }

  pass = 1;

err:
  if (buf != NULL) free(buf);

  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_capture_image_quick_timeout_no_change(void)
{
  int pass;
  unsigned char *buf;
  size_t buf_size;
  int status;
  epicsTimeStamp start_time;
  epicsTimeStamp now_time;
  double elapsed_time;

  pass = 0;
  buf = NULL;
  buf_size = CMAD_TEST_IMAGE_WIDTH * CMAD_TEST_IMAGE_HEIGHT * 3 * sizeof(*buf);

  cmad_test_set_up();

  status = cmad_start_image_capture(client);
  if (status == -1) {
    testDiag("unable to start image capture");
    goto err;
  }

  buf = malloc(buf_size);
  if (buf == NULL) {
    testDiag("unable to allocate image buffer");
    goto err;
  }

  cmad_test_set_acquire_period(2.0);
  cmad_test_ca_worker_do();

  /* Synchronize with the 2.0 second acquire period. */
  for (;;) {
    status = epicsTimeGetCurrent(&start_time);
    if (status == -1) {
      testDiag("unable to obtain current time");
      goto err;
    }
    status = cmad_capture_image(client, buf, buf_size);
    if (status != 0) {
      testDiag("image capture should have succeeded");
      goto err;
    }
    status = epicsTimeGetCurrent(&now_time);
    if (status == -1) {
      testDiag("unable to obtain current time");
      goto err;
    }
    elapsed_time = epicsTimeDiffInSeconds(&now_time, &start_time);
    if (elapsed_time < 0.0) {
      /* Time advanced backward; consider now to be the new start time. */
      start_time = now_time;
    } else if (elapsed_time > 1.75) {
      /* Synchronization achieved. */
      break;
    }
  }

  status = cmad_set_timeout(client, 1.0);
  if (status == -1) {
    testDiag("unable to set timeout");
    goto err;
  }

  /* A quick capture does not wait for a new image to arrive; it just takes the
   * most recent one, so it should return successfully right away. */
  status = cmad_capture_image_quick(client, buf, buf_size);
  if (status != 0) {
    testDiag("image capture timed out");
    goto err;
  }

  pass = 1;

err:
  if (buf != NULL) free(buf);

  cmad_test_tear_down();

  return pass;
}

static int
cmad_test_test_capture_image_quick_multiple_clients(void)
{
  int i;
  int j;
  int pass;
  unsigned char *buf;
  size_t buf_size;
  struct cmad_client *client2;
  int status;

  i = 0;
  j = 0;
  pass = 0;
  buf = NULL;
  buf_size = CMAD_TEST_IMAGE_WIDTH * CMAD_TEST_IMAGE_HEIGHT * 3 * sizeof(*buf);
  client2 = NULL;

  cmad_test_set_up();

  client2 = cmad_client_new(CMAD_TEST_PV_PREFIX, CMAD_TEST_IMAGE_WIDTH,
      CMAD_TEST_IMAGE_HEIGHT, CMAD_TEST_TIMEOUT);
  if (client2 == NULL) {
    testDiag("unable to create client2");
    goto err;
  }

  for (j = 0; j < 2; j++) {
    status = cmad_start_image_capture((j == 0) ? client : client2);
    if (status == -1) {
      testDiag("unable to start client%s image capture", (j == 0) ? "" : "2");
      goto err;
    }
  }

  buf = malloc(buf_size);
  if (buf == NULL) {
    testDiag("unable to allocate image buffer");
    goto err;
  }

  for (i = 0; i < 4; i++) {
    for (j = 0; j < 2; j++) {
      status = cmad_capture_image_quick((j == 0) ? client : client2, buf,
          buf_size);
      if (status == -1) {
        testDiag("unable to capture client%s image: %s", (j == 0) ? "" : "2",
            cmad_error(client));
        goto err;
      }
    }
  }

  pass = 1;

err:
  if (buf != NULL) free(buf);
  if (client2 != NULL) cmad_client_free(client2);

  cmad_test_tear_down();

  return pass;
}

static void
cmad_test_program_set_up(void)
{
  client = NULL;

  ca_worker_do_eid = epicsEventMustCreate(epicsEventEmpty);
  ca_worker_ack_eid = epicsEventMustCreate(epicsEventEmpty);

  ca_worker_mid = epicsMutexMustCreate();
  ca_worker_stop = 0;
  ca_worker_value[CMAD_TEST_CH_ACQUIRE_PERIOD] = &ca_worker_acquire_period;
  ca_worker_value[CMAD_TEST_CH_ACQUIRE] = &ca_worker_acquire;
  ca_worker_value[CMAD_TEST_CH_ARRAY_CALLBACKS] = &ca_worker_array_callbacks;
  ca_worker_value[CMAD_TEST_CH_SIZE_X] = &ca_worker_size_x;
  ca_worker_value[CMAD_TEST_CH_SIZE_Y] = &ca_worker_size_y;
  ca_worker_value[CMAD_TEST_CH_COLOR_MODE] = &ca_worker_color_mode;
  ca_worker_value[CMAD_TEST_CH_IMAGE1_ENABLE] = &ca_worker_image1_enable;
  cmad_test_reset_ca_worker_communication_locked();

  (void)epicsThreadCreate("cmad CA worker", epicsThreadPriorityLow,
      epicsThreadGetStackSize(epicsThreadStackSmall), cmad_test_ca_worker,
      NULL);
}

static void
cmad_test_program_tear_down(void)
{
  cmad_test_reset_ca_worker_communication();
  cmad_test_set_ca_worker_stop(1);
  cmad_test_ca_worker_do();

  epicsEventDestroy(ca_worker_do_eid);
  epicsEventDestroy(ca_worker_ack_eid);

  epicsMutexDestroy(ca_worker_mid);
}

static void
cmad_test_set_up(void)
{
  cmad_test_set_up_with_timeout(CMAD_TEST_TIMEOUT);
}

static void
cmad_test_set_up_with_timeout(double client_timeout)
{
  cmad_test_reset_ca_worker_communication();
  cmad_test_set_defaults();
  cmad_test_ca_worker_do();
  cmad_test_reset_ca_worker_communication();

  client = cmad_client_new(CMAD_TEST_PV_PREFIX, CMAD_TEST_IMAGE_WIDTH,
      CMAD_TEST_IMAGE_HEIGHT, client_timeout);
}

static void
cmad_test_tear_down(void)
{
  cmad_client_free(client);
  client = NULL;

  cmad_test_reset_ca_worker_communication();
  cmad_test_set_acquire(0);
  cmad_test_ca_worker_do();
}

static void
cmad_test_reset_ca_worker_communication(void)
{
  epicsMutexMustLock(ca_worker_mid);

  cmad_test_reset_ca_worker_communication_locked();

  epicsMutexUnlock(ca_worker_mid);
}

static void
cmad_test_reset_ca_worker_communication_locked(void)
{
  int i;

  for (i = 0; i < CMAD_TEST_NUM_CHANNELS; i++) {
    ca_worker_marked[i] = 0;
  }
  ca_worker_err_msg[0] = '\0';
}

static void
cmad_test_set_ca_worker_stop(int value)
{
  epicsMutexMustLock(ca_worker_mid);

  ca_worker_stop = value;

  epicsMutexUnlock(ca_worker_mid);
}

static void
cmad_test_set_acquire_period(dbr_double_t value)
{
  epicsMutexMustLock(ca_worker_mid);

  ca_worker_acquire_period = value;
  ca_worker_marked[CMAD_TEST_CH_ACQUIRE_PERIOD] = 1;

  epicsMutexUnlock(ca_worker_mid);
}

static void
cmad_test_set_acquire(dbr_enum_t value)
{
  epicsMutexMustLock(ca_worker_mid);

  ca_worker_acquire = value;
  ca_worker_marked[CMAD_TEST_CH_ACQUIRE] = 1;

  epicsMutexUnlock(ca_worker_mid);
}

static void
cmad_test_set_array_callbacks(dbr_enum_t value)
{
  epicsMutexMustLock(ca_worker_mid);

  ca_worker_array_callbacks = value;
  ca_worker_marked[CMAD_TEST_CH_ARRAY_CALLBACKS] = 1;

  epicsMutexUnlock(ca_worker_mid);
}

static void
cmad_test_set_size_x(dbr_long_t value)
{
  epicsMutexMustLock(ca_worker_mid);

  ca_worker_size_x = value;
  ca_worker_marked[CMAD_TEST_CH_SIZE_X] = 1;

  epicsMutexUnlock(ca_worker_mid);
}

static void
cmad_test_set_size_y(dbr_long_t value)
{
  epicsMutexMustLock(ca_worker_mid);

  ca_worker_size_y = value;
  ca_worker_marked[CMAD_TEST_CH_SIZE_Y] = 1;

  epicsMutexUnlock(ca_worker_mid);
}

static void
cmad_test_set_color_mode(dbr_enum_t value)
{
  epicsMutexMustLock(ca_worker_mid);

  ca_worker_color_mode = value;
  ca_worker_marked[CMAD_TEST_CH_COLOR_MODE] = 1;

  epicsMutexUnlock(ca_worker_mid);
}

static void
cmad_test_set_image1_enable(dbr_enum_t value)
{
  epicsMutexMustLock(ca_worker_mid);

  ca_worker_image1_enable = value;
  ca_worker_marked[CMAD_TEST_CH_IMAGE1_ENABLE] = 1;

  epicsMutexUnlock(ca_worker_mid);
}

static void
cmad_test_set_defaults(void)
{
  cmad_test_set_acquire_period(1.0 / CMAD_TEST_FPS);
  cmad_test_set_acquire(1);
  cmad_test_set_array_callbacks(1);
  cmad_test_set_size_x(CMAD_TEST_IMAGE_WIDTH);
  cmad_test_set_size_y(CMAD_TEST_IMAGE_HEIGHT);
  cmad_test_set_color_mode(CMAD_TEST_COLOR_MODE);
  cmad_test_set_image1_enable(1);
}

static int
cmad_test_wait_connected(struct cmad_client *c, int connected, double timeout)
{
  int status;
  epicsTimeStamp start_time;
  epicsTimeStamp now_time;
  double elapsed_time;

  status = epicsTimeGetCurrent(&start_time);
  if (status == -1) {
    testAbort("unable to obtain current time");
    return -1;
  }

  for (;;) {
    if (cmad_is_connected(c) == connected) return 0;
    status = epicsTimeGetCurrent(&now_time);
    if (status == -1) {
      testAbort("unable to obtain current time");
      return -1;
    }
    elapsed_time = epicsTimeDiffInSeconds(&now_time, &start_time);
    if (elapsed_time < 0.0) {
      /* Time advanced backward; consider now to be the new start time. */
      start_time = now_time;
    } else if (elapsed_time > timeout) {
      return -1;
    }
    epicsThreadSleep(0.1);
  }
}

static int
cmad_test_wait_ready(struct cmad_client *c, int ready, double timeout)
{
  int status;
  epicsTimeStamp start_time;
  epicsTimeStamp now_time;
  double elapsed_time;

  status = epicsTimeGetCurrent(&start_time);
  if (status == -1) {
    testAbort("unable to obtain current time");
    return -1;
  }

  for (;;) {
    if (cmad_is_ready(c) == ready) return 0;
    status = epicsTimeGetCurrent(&now_time);
    if (status == -1) {
      testAbort("unable to obtain current time");
      return -1;
    }
    elapsed_time = epicsTimeDiffInSeconds(&now_time, &start_time);
    if (elapsed_time < 0.0) {
      /* Time advanced backward; consider now to be the new start time. */
      start_time = now_time;
    } else if (elapsed_time > timeout) {
      return -1;
    }
    epicsThreadSleep(0.1);
  }
}

static void
cmad_test_ca_worker_do(void)
{
  int ok;
  epicsEventWaitStatus event_status;

  ok = 0;

  epicsEventSignal(ca_worker_do_eid);
  event_status = epicsEventWaitWithTimeout(ca_worker_ack_eid,
      CMAD_TEST_TIMEOUT);
  if (event_status != epicsEventWaitOK) {
    testAbort("%s out waiting for CA worker to signal ACK",
        (event_status == epicsEventWaitTimeout) ? "timed" : "errored");
  } else {
    epicsMutexMustLock(ca_worker_mid);
    ok = (ca_worker_err_msg[0] == '\0') ? 1 : 0;
    if (!ok) {
      testDiag("CA worker error: %s", ca_worker_err_msg);
    }
    epicsMutexUnlock(ca_worker_mid);
    if (!ok) {
      testAbort("CA worker failed");
    }
  }
}

static void
cmad_test_ca_worker(void *param)
{
  int locked;
  int signaled;
  int context_created;
  int i;
  const char *pv_name[CMAD_TEST_NUM_CHANNELS];
  chtype ch_type[CMAD_TEST_NUM_CHANNELS];
  chid ch[CMAD_TEST_NUM_CHANNELS];
  int ch_created[CMAD_TEST_NUM_CHANNELS];
  int status;
  char err_msg[CMAD_TEST_CA_WORKER_ERR_MSG_SIZE];
  epicsEventWaitStatus event_status;

  locked = 0;
  signaled = 0;
  context_created = 0;
  err_msg[0] = '\0';

  i = CMAD_TEST_CH_ACQUIRE_PERIOD;
  pv_name[i] = AD_PV_NAME_ACQUIRE_PERIOD;
  ch_type[i] = DBR_DOUBLE;
  ch_created[i] = 0;

  i = CMAD_TEST_CH_ACQUIRE;
  pv_name[i] = AD_PV_NAME_ACQUIRE;
  ch_type[i] = DBR_ENUM;
  ch_created[i] = 0;

  i = CMAD_TEST_CH_ARRAY_CALLBACKS;
  pv_name[i] = AD_PV_NAME_ARRAY_CALLBACKS;
  ch_type[i] = DBR_ENUM;
  ch_created[i] = 0;

  i = CMAD_TEST_CH_SIZE_X;
  pv_name[i] = AD_PV_NAME_SIZE_X;
  ch_type[i] = DBR_LONG;
  ch_created[i] = 0;

  i = CMAD_TEST_CH_SIZE_Y;
  pv_name[i] = AD_PV_NAME_SIZE_Y;
  ch_type[i] = DBR_LONG;
  ch_created[i] = 0;

  i = CMAD_TEST_CH_COLOR_MODE;
  pv_name[i] = AD_PV_NAME_COLOR_MODE;
  ch_type[i] = DBR_ENUM;
  ch_created[i] = 0;

  i = CMAD_TEST_CH_IMAGE1_ENABLE;
  pv_name[i] = AD_PV_NAME_IMAGE1_ENABLE;
  ch_type[i] = DBR_ENUM;
  ch_created[i] = 0;

  status = ca_context_create(ca_disable_preemptive_callback);
  if (status != ECA_NORMAL) {
    (void)epicsSnprintf(err_msg, sizeof(err_msg),
        "unable to create CA context: %s", ca_message(status));
  }
  context_created = 1;

  if (err_msg[0] == '\0') {
    for (i = 0; i < CMAD_TEST_NUM_CHANNELS; i++) {
      status = ca_create_channel(pv_name[i], NULL, NULL, 0, &ch[i]);
      if (status != ECA_NORMAL) {
        (void)epicsSnprintf(err_msg, sizeof(err_msg),
            "unable to create CA channel \"%s\": %s", pv_name[i],
            ca_message(status));
        break;
      }
      ch_created[i] = 1;
    }
  }

  if (err_msg[0] == '\0') {
    status = ca_pend_io(CMAD_TEST_TIMEOUT);
    if (status != ECA_NORMAL) {
      (void)epicsSnprintf(err_msg, sizeof(err_msg),
          "not all CA channels connected within %.1f seconds: %s",
          CMAD_TEST_TIMEOUT, ca_message(status));
    }
  }

  for (;;) {
    event_status = epicsEventWaitWithTimeout(ca_worker_do_eid, 1.0);
    if (event_status != epicsEventWaitOK) {
      if (event_status == epicsEventWaitTimeout) {
        status = ca_poll();
        if (status != ECA_TIMEOUT) {
          if (err_msg[0] == '\0') {
            (void)epicsSnprintf(err_msg, sizeof(err_msg),
                "unable to poll CA client library: %s", ca_message(status));
          }
          goto err;
        }
        continue;
      } else {
        if (err_msg[0] == '\0') {
          (void)epicsSnprintf(err_msg, sizeof(err_msg),
              "unable to wait for \"do\" event");
        }
        goto err;
      }
    }
    signaled = 1;

    if (err_msg[0] != '\0') goto err;

    epicsMutexMustLock(ca_worker_mid);
    locked = 1;

    if (ca_worker_stop) goto out;

    for (i = 0; i < CMAD_TEST_NUM_CHANNELS; i++) {
      if (!ca_worker_marked[i]) continue;
      status = ca_put(ch_type[i], ch[i], ca_worker_value[i]);
      if (status != ECA_NORMAL) {
        (void)epicsSnprintf(err_msg, sizeof(err_msg),
            "unable to perform CA put to \"%s\": %s", pv_name[i],
            ca_message(status));
        goto err;
      }
    }

    status = ca_pend_io(CMAD_TEST_TIMEOUT);
    if (status != ECA_NORMAL) {
      (void)epicsSnprintf(err_msg, sizeof(err_msg),
          "unable to perform CA put(s): %s", ca_message(status));
      goto err;
    }

    epicsMutexUnlock(ca_worker_mid);
    locked = 0;

    epicsEventSignal(ca_worker_ack_eid);
    signaled = 0;
  }

  goto out;

err:
    if (signaled && err_msg[0] != '\0') {
      if (!locked) {
        epicsMutexMustLock(ca_worker_mid);
        locked = 1;
      }

      (void)epicsSnprintf(ca_worker_err_msg, sizeof(ca_worker_err_msg), "%s",
          err_msg);
    }

out:
    if (locked) epicsMutexUnlock(ca_worker_mid);

    if (signaled) epicsEventSignal(ca_worker_ack_eid);

    for (i = 0; i < CMAD_TEST_NUM_CHANNELS; i++) {
      if (ch_created[i]) (void)ca_clear_channel(ch[i]);
    }
    for (i = 0; i < CMAD_TEST_NUM_CHANNELS; i++) {
      if (ch_created[i]) {
        (void)ca_flush_io();
        break;
      }
    }
    if (context_created) ca_context_destroy();
}
